from splintr.modelparallel.basic_model_parallel import BasicModelParallel

""" ModelParallel returns appropriate ModelParallel training option based on user
input. Currently Supports:

    -- 'basic': Basic Model Parallelism """

def ModelParallel(**kwargs):

    if kwargs['style'] == 'basic':
        model = kwargs['model']
        layer_split = kwargs['layer_split']

        return BasicModelParallel(model, layer_split)

    else:
        raise NotImplementedError("No other Model Parallel training methods have been implemented yet!")



__all__ = ['DataParallel', 'ModelParallel', 'benchmarks', 'dataparallel', 'modelparallel']
from .DataParallel import DataParallel
from .ModelParallel import ModelParallel
import splintr.benchmarks
import splintr.dataparallel
import splintr.modelparallel



import torch
import torch.nn as nn

""" VGG11 Stucture for arbitrary input size: https://arxiv.org/pdf/1409.1556.pdf,
    implementation modelled after torchvision """
    
class VGG11(nn.Module):

    def __init__(self, num_classes):
        super(VGG11, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace = True),
            nn.MaxPool2d(kernel_size = 2, stride = 2),
            nn.Conv2d(64, 128, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace = True),
            nn.MaxPool2d(kernel_size = 2, stride = 2),
            nn.Conv2d(128, 256, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace = True),
            nn.Conv2d(256, 256, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace = True),
            nn.MaxPool2d(kernel_size = 2, stride = 2),
            nn.Conv2d(256, 512, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace = True),
            nn.Conv2d(512, 512, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace = True),
            nn.MaxPool2d(kernel_size = 2, stride = 2),
            nn.Conv2d(512, 512, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace = True),
            nn.Conv2d(512, 512, kernel_size = 3, padding = 1),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace = True),
            nn.MaxPool2d(kernel_size = 2, stride = 2),
            nn.AdaptiveAvgPool2d((7, 7)),
            nn.Flatten(),
            nn.Linear(512 * 7 * 7, 4096, bias = True),
            nn.ReLU(inplace = True),
            nn.Dropout(p = 0.5),
            nn.Linear(4096, 4096, bias = True),
            nn.ReLU(inplace = True),
            nn.Dropout(p = 0.5),
            nn.Linear(4096, num_classes, bias = True))

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode = 'fan_out', nonlinearity = 'relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        return self.features(x)



import torch.nn as nn
import splintr.dataparallel.isubnet as isub
import splintr.dataparallel.local_sgd as lsgd


""" DataParallel returns appropriate DataParallel training option based on user
input. Currently Supports:

	-- 'torch': PyTorch style Data Parallelism
	-- 'localSGD': Vanilla Local SGD (arXiv:1808.07217)
	-- 'ist': Indepent Subnet Training W/ Vanilla Local SGD (arXiv:1910.02120) """

def DataParallel(**kwargs):

	if kwargs['style'] == 'torch':
		model = kwargs['model']
		device_ids = kwargs['device_ids'] if 'device_ids' in kwargs else None
		output_device = kwargs['output_device'] if 'output_device' in kwargs else None
		dim = kwargs['dim'] if 'dim' in kwargs else 0

		return nn.DataParallel(model.cuda(), device_ids = device_ids, output_device = output_device, dim = dim)

	elif kwargs['style'] == 'localSGD':
		model = kwargs['model']
		iterations = kwargs['iterations'] if 'iterations' in kwargs else 40
		initializer = kwargs['initializer'] if 'initializer' in kwargs else lambda x: x

		return lsgd.LocalSGD(model, iterations = iterations, initializer = initializer)

	elif kwargs['style'] == 'ist':
		model = kwargs['model']
		iterations = kwargs['iterations'] if 'iterations' in kwargs else 40
		initializer = kwargs['initializer'] if 'initializer' in kwargs else lambda x: x

		return isub.ISubnet(model, iterations = iterations, initializer = initializer)

	else:
		raise NotImplementedError("No other Data Parallel training methods have been implemented yet!")



import torch
import torch.nn as nn

""" Class BasicModelParallel represents a "vanilla" model-parallelized version
of the input model. In this case we simply split the original model across GPUs
with not attempt to implement more exotic model-parallel styles like input
pipelining etc. 

    -- model (nn.Module / nn.Sequential): input model
    -- layer_split (list of ints): index of layer splits for the model. Example:
    [16] means to split layers [0, 15] to cuda:0 and layers [16,] to cuda:1 """

class BasicModelParallel(nn.Module):

    def __init__(self, model, layer_split):
        super(BasicModelParallel, self).__init__()

        assert len(layer_split) == torch.cuda.device_count() - 1

        """ iterating through model features currently requires that input model 
        has an nn.Sequential self.features variable that presents the model 
        architecture or is an nn.Sequential """

        if (not isinstance(model, nn.Sequential)):
            model = model.features

        modules = [i[1] for i in list(model._modules.items())]
        layer_split = [0] + layer_split + [len(modules)]

        self.sub_models = []

        for i in range(len(layer_split) - 1):
            self.sub_models.append(nn.Sequential(*modules[layer_split[i]:layer_split[i + 1]]).cuda(i))

    def forward(self, x):
        for i in range(len(self.sub_models)):
            x = self.sub_models[i](x.cuda(i))

        return x.cpu()

    def parameters(self):
        params = []

        for i in self.sub_models:
            params = params + list(i.parameters())

        return params

    def eval(self):
        for i in self.sub_models:
            i.training = False

    def train(self):
        for i in self.sub_models:
            i.training = True



import os
import pickle
import tarfile
import requests
import numpy as np


""" Returns Generators For Training And Testing CIFAR-100 Data. 
    Generators Produce Tuples Of Image (32, 32, 3) And One-Hot Encoded
    Labels. """

def cifar100():

    url = 'https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz'
    open('cifar-100.tar.gz', "wb").write(requests.get(url).content)

    tar_out = tarfile.open('cifar-100.tar.gz', 'r:gz')

    for member in tar_out.getmembers():
        if (member.name == 'cifar-100-python/train'):
            train = pickle.load(tar_out.extractfile(member), encoding = 'bytes')

        elif (member.name == 'cifar-100-python/test'):
            test = pickle.load(tar_out.extractfile(member), encoding = 'bytes')

    def data_generator(data):
        for i in range(len(data[b'data'])):
            image = data[b'data'][i]
            label = data[b'fine_labels'][i]

            image = image.reshape((3, 1024))
            out_image = []

            for j in range(32):
                row = []

                for k in range(32):
                    row.append([image[0][j * 32 + k], image[1][j * 32 + k], image[2][j * 32 + k]])

                out_image.append(row)

            yield (np.array(out_image), label)

    os.remove('cifar-100.tar.gz')

    return (data_generator(train), data_generator(test))



import copy
import torch
import threading
import torch.nn as nn

""" Class LocalSGD implements Vanilla Local SGD as described in 
(arXiv:1808.07217) for multi-GPU training of sequential models.

    -- model (nn.Sequential): input model for this component
    -- iterations (int): local SGD iterations between merging updates
    -- initializer (function): Weight initializer. During benchmarking, we found
    that separately initializing the model replicas weight produced better
    results """   

class LocalSGD(nn.Module):

    def __init__(self, model, iterations = 40, initializer = lambda x: x):
        super(LocalSGD, self).__init__()
        self.split = torch.cuda.device_count()
        self.broadcast_map = dict()
        self.iterations = iterations
        self.iteration_count = 0

        """ broadcast_map maintains state of model instances across devices """

        for i in range(self.split):
            self.broadcast_map[i] = copy.deepcopy(model).cuda(i)
            initializer(self.broadcast_map[i])

        self.params = []

        for i in range(self.split):
            self.params.append(list(self.broadcast_map[i].parameters()))

    def forward(self, x):

        if self.training:
            if self.iteration_count == self.iterations:
                self.reset()
                self.iteration_count = 0

            x = torch.cuda.comm.scatter(x, devices = list(range(self.split)))
            outputs = [None] * self.split

            def thread(inp, num):
                outputs[num] = self.broadcast_map[num](inp)

            threads = [threading.Thread(target = thread, args = (x[i], i)) for i in range(self.split)]

            for t in threads:
                t.start()

            for t in threads:
                t.join()

            self.iteration_count += 1
                
            return torch.cuda.comm.gather(outputs, destination = -1)
        else:
            x = x.cuda(0)
            return self.broadcast_map[0](x).cpu()

    def parameters(self):
        params = []

        for i in self.params:
            params = params + i

        return params

    """ Reset performs a local-SGD iteration across model instances """

    def reset(self):
        param_data = None

        for i in range(len(self.params[0])):
            for j in range(self.split):
                if param_data != None:
                    param_data.data = param_data.data + self.params[j][i].data.clone().cpu()
                else:
                    param_data = self.params[j][i].data.clone().cpu()
                    
            param_data.data = param_data.data / self.split
            
            for j in range(self.split):
                self.params[j][i].data = param_data.data.clone().cuda(j)
                
            param_data = None


import setuptools

setuptools.setup (name = 'splintr',
        version = '0.0.5',
        description = 'One-Line Single-Machine Multi-GPU DNN Training',
        url = 'https://github.com/splintr-ml/splintr',
        author = 'splintr-ml',
        author_email = 'splintrml@gmail.com',
        install_requires = ['torch', 'numpy'],
        packages = setuptools.find_packages(),
        classifiers = ['Development Status :: 3 - Alpha', 'Environment :: GPU', 'Programming Language :: Python :: 3'])



import copy
import torch
import threading
import torch.nn as nn
import splintr.dataparallel.local_sgd as lsgd

""" We implement multi-GPU indepent-subnet training here as tt is described in:
    arXiv:1910.02120. """

""" Class ISubnetFC constitutes the independently trained subcomponents of the
original model's fully-connected components.

    -- in_features (int): number of features in the input Tensor
    -- out_features (int): desired number of features in the output Tensor
    -- back (ISubnetFC): internal variable to map ISubnetFC layers before this
    -- activation (nn.Module / nn.Functional): activation function """

class ISubnetFC(nn.Module):

    def __init__(self, in_features, out_features, back, activation = None):
        super(ISubnetFC, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

        # "Master" Weight, utilized for evaluation, etc.

        self.weight = torch.zeros(out_features, in_features)
        nn.init.normal_(self.weight, 0, 0.01)

        # Correct for distributional shift referenced in paper

        self.norm = nn.BatchNorm1d(num_features = out_features)

        self.back = back
        self.final = False

        self.activation = activation if activation else lambda x: x

        """ Sets up initialization of masks, sync's input mask with previous output
        mask """

        self.input_mask = torch.ones(in_features) if not self.back else self.back.mask_sync()
        self.output_mask = torch.randint(low = 0, high = torch.cuda.device_count(), size = (out_features,))

        self.split = torch.cuda.device_count()

        """ mask_map maintains state information about weights and masks across
        devices """

        self.mask_map = dict()
        self.params = [list(self.norm.parameters())]

        for i in range(self.split):
            weight = self.weight.clone().detach().cuda(i)
            weight.requires_grad = True

            if self.back:
                self.mask_map[i] = (weight, (self.input_mask == i).clone().cuda(i), (self.output_mask == i).clone().cuda(i))
            else:
                self.mask_map[i] = (weight, self.input_mask.clone().cuda(i), (self.output_mask == i).clone().cuda(i))

            self.params.append([weight])

    def forward(self, x):
        if self.training:
            x = torch.cuda.comm.scatter(x, devices = range(self.split))
            outputs = [None] * self.split

            def thread(inp, num):
                weight, input_mask, output_mask = self.mask_map[num]
                outputs[num] = (input_mask * inp).matmul(weight.t()) * output_mask

            threads = [threading.Thread(target = thread, args = (x[i], i)) for i in range(self.split)]

            for t in threads:
                t.start()

            for t in threads:
                t.join()

            return self.activation(self.norm(torch.cuda.comm.gather(outputs, destination = -1)))
        else:
            return self.activation(self.norm(x.matmul(self.weight.t())))

    def parameters(self):
        params = []

        for i in self.params:
            params = params + i

        return params

    """ "Resets" the masks and "master" weights to propogate updates and start 
    next "cycle" of masking """

    def reset(self):
        index_masks = []

        for i in range(self.split):
            _, input_mask, output_mask = self.mask_map[i]
            output_mask = output_mask.resize(self.out_features, 1)
            index_masks.append(input_mask * output_mask)

        for i in range(self.split):
            self.weight.data[index_masks[i].bool().cpu()] = self.mask_map[i][0].data.clone()[index_masks[i].bool()].cpu()

        self.input_mask = torch.ones(self.in_features) if not self.back else self.back.mask_sync()
        self.output_mask = torch.randint(low = 0, high = torch.cuda.device_count(), size = (self.out_features,)) if not self.final else torch.ones(self.out_features)

        for i in range(self.split):
            self.mask_map[i][0].data = self.weight.data.clone().cuda(i)

            input_mask = (self.input_mask == i).clone().cuda(i) if self.back else self.input_mask.clone().cuda(i)
            output_mask = (self.output_mask == i).clone().cuda(i) if not self.final else self.output_mask.clone().cuda(i)

            self.mask_map[i] = (self.mask_map[i][0], input_mask, output_mask)

    """ Sets up "forward" variable for any "ahead" ISubnetFC layers for mask
    syncing """

    def add_forw(self, forw):
        self.forw = forw

    """ Serves output_mask to sync with self.forw """

    def mask_sync(self):
        return self.output_mask

    """ Designates this ISubnetFC as last in chain so that output_mask is 
    configured correctly """

    def make_final(self):
        self.final = True
        self.output_mask = torch.ones(self.out_features)

        for i in range(self.split):
            self.mask_map[i] = (self.mask_map[i][0], self.mask_map[i][1], self.output_mask.clone().cuda(i))
            

""" Class ISubnet represents the model class that is returned to the user. It
separates the input model appropriately and acts as the interface between the
corresponding ISubnetNFC and ISubnetFC instances. 

    -- model (nn.Module / nn.Sequential): input model for this component
    -- iterations (int): number of iterations before merging weights in local SGD
    and between remasking for the fully-connected layers
    -- initializer (function): Weight initializer. During benchmarking, we found
    that separately initializing the model replicas weight produced better
    results """

class ISubnet(nn.Module):

    def __init__(self, model, iterations = 40, initializer = lambda x: x):
        super(ISubnet, self).__init__()

        self.iterations = iterations
        self.iteration_count = 0
        self.non_subnet_features = []
        self.subnet_features = []

        found_linear = False
        curr_linear = None
        back_linear = None
        fc = None

        """ iterating through model features to find non-subnet features and 
        subnet features currently requires that input model has an nn.Sequential
        self.features variable that presents the model architecture or is an 
        nn.Sequential """

        if (not isinstance(model, nn.Sequential)):
            model = model.features

        for _, layer in model._modules.items():

            if not found_linear:
                if isinstance(layer, nn.Linear):
                    found_linear = True
                    curr_linear = layer

                else:
                    self.non_subnet_features.append(layer)

            else:
                if curr_linear:

                    if isinstance(layer, nn.Linear):
                        fc = ISubnetFC(curr_linear.in_features, curr_linear.out_features, back_linear)
                        curr_linear = layer

                    else:
                        fc = ISubnetFC(curr_linear.in_features, curr_linear.out_features, back_linear, layer)
                        curr_linear = None

                    if back_linear:
                        back_linear.add_forw(fc)

                    self.subnet_features.append(fc)
                    back_linear = fc

                else:
                    if isinstance(layer, nn.Linear):
                        curr_linear = layer

        if curr_linear:
            fc = ISubnetFC(curr_linear.in_features, curr_linear.out_features, back_linear)

            if back_linear:
                back_linear.add_forw(fc)

            self.subnet_features.append(fc)

        if fc:
            fc.make_final()

        self.non_subnet_features = lsgd.LocalSGD(model = nn.Sequential(*self.non_subnet_features), iterations = self.iterations, initializer = initializer)

    def forward(self, x):

        if self.training:
            if self.iteration_count == self.iterations:
                self.reset()
                self.iteration_count = 0

            x = self.non_subnet_features(x)
            for i in self.subnet_features:
                x = i(x)

            self.iteration_count += 1

            return x
        else:
            with torch.no_grad():
                x = self.non_subnet_features(x)
                for i in self.subnet_features:
                    x = i(x)

                return x

    def parameters(self):
        params = self.non_subnet_features.parameters()

        for i in self.subnet_features:
            params = params + i.parameters()

        return params

    """ Manually setting up eval and train functions for the non-subnet and 
    subnet components as well as normalization layers in subnet components """

    def eval(self):
        self.training = False
        self.non_subnet_features.training = False

        for i in self.subnet_features:
            i.training = False
            i.norm.training = False

    def train(self):
        self.training = True
        self.non_subnet_features.training = True

        for i in self.subnet_features:
            i.training = True
            i.norm.training = True

    """ Calls reset on the sub-components, called after specified number of 
    iterations """

    def reset(self):
        for i in self.subnet_features:
            i.reset()



import time
import torch
import numpy as np
import torch.nn as nn
import torch.optim as optim
import splintr.benchmarks.data_loader as dl

""" CIFAR100 Benchmark """

def cifar100(model, wrapper = None, batch_size = 128, target_acc = 1, target_top10 = 1, evaluate = True, verbose = 3):

    train, test = dl.cifar100()
    train, test = list(train), list(test)

    if verbose:
        print('--- Data Loaded ---')

    device = torch.device('cpu')

    if torch.cuda.is_available() and not wrapper:
        device = torch.device('cuda')

    if wrapper:
        model = wrapper(model)
    else:
        model = model.to(device)

    train_images = torch.from_numpy(np.array([np.moveaxis(i[0], 2, 0) for i in train])).float()
    test_images = torch.from_numpy(np.array([np.moveaxis(i[0], 2, 0) for i in test])).float()

    if not wrapper:
        train_images = train_images.to(device)
        test_images = test_images.to(device)

    loss_func = nn.CrossEntropyLoss()
    optimizer = optim.SGD(model.parameters(), lr = 0.01)

    tot_time = 0
    num_mb_epoch = int((50000 / batch_size)) + 1
    mb_left = 30 * num_mb_epoch
    for epoch in range(30):
        avg_batch_time = 0
        loss_over_minibatch = 0
        prev_time = time.time()

        for mb in range(num_mb_epoch):
            inputs = train_images[mb * batch_size:(mb + 1) * batch_size]
            labels = torch.from_numpy(np.array([i[1] for i in train[mb * batch_size:(mb + 1) * batch_size]])).long()

            if not wrapper:
                labels = labels.to(device)

            optimizer.zero_grad()

            outputs = model(inputs)

            loss = loss_func(outputs, labels.to(outputs.device))
            loss.backward()
            optimizer.step()

            if torch.cuda.is_available():
                for i in range(torch.cuda.device_count()):
                    torch.cuda.synchronize(torch.device('cuda', i))

            tot_time += time.time() - prev_time
            avg_batch_time += time.time() - prev_time
            loss_over_minibatch += loss.item()
            prev_time = time.time()
            mb_left -= 1

            if (not wrapper) and (torch.cuda.is_available()):
                del inputs
                del labels
                del outputs
                torch.cuda.empty_cache()

            if ((mb % 50) == 49) and (verbose == 3) and evaluate:
                print('Epoch ' + str(epoch + 1) + ' Loss: ' + str(loss_over_minibatch / 50))
                print('Time Per Batch: ' + str(avg_batch_time / 50) + 's Time Left: ' + str((avg_batch_time / 50) * mb_left) + 's')

                inputs = test_images
                labels = torch.from_numpy(np.array([i[1] for i in test])).long()

                if not wrapper:
                    labels = labels.to(device)

                model.eval()
                with torch.no_grad():
                    outputs = model(inputs)
                model.train()

                _, predicted = torch.max(outputs, 1)
                _, top10 = torch.topk(outputs, 10, dim = 1)

                predicted = predicted.to(labels.device)
                top10 = top10.to(labels.device)

                corr = 0
                top10_corr = 0

                for i in range(len(predicted)):
                    if predicted[i].item() == labels[i]:
                        corr += 1
                    if labels[i] in top10[i]:
                        top10_corr += 1

                print('Acc: ' + str(corr / len(predicted)))
                print('Top-10: ' + str(top10_corr / len(predicted)))

                if ((corr / len(predicted)) >= target_acc) and ((top10_corr / len(predicted)) >= target_top10):
                    print("Total Time: " + str(tot_time) + "s")
                    return

                if not wrapper and torch.cuda.is_available():
                    del inputs
                    del labels
                    del outputs
                    del predicted
                    del top10
                    torch.cuda.empty_cache()

                if torch.cuda.is_available():
                    for i in range(torch.cuda.device_count()):
                        torch.cuda.synchronize(torch.device('cuda', i))

                avg_batch_time = 0
                loss_over_minibatch = 0
                prev_time = time.time()

    if (verbose >= 2) and evaluate:
        inputs = test_images
        labels = torch.from_numpy(np.array([i[1] for i in test])).long()

        if not wrapper:
            labels = labels.to(device)

        model.eval()
        with torch.no_grad():
            outputs = model(inputs)
        model.train()

        _, predicted = torch.max(outputs, 1)
        _, top10 = torch.topk(outputs, 10, dim = 1)

        predicted = predicted.to(labels.device)
        top10 = top10.to(labels.device)

        corr = 0
        top10_corr = 0

        for i in range(len(predicted)):
            if predicted[i].item() == labels[i]:
                corr += 1
            if labels[i] in top10[i]:
                top10_corr += 1

        print('Acc: ' + str(corr / len(predicted)))
        print('Top-10: ' + str(top10_corr / len(predicted)))

        if ((corr / len(predicted)) >= target_acc) and ((top10_corr / len(predicted)) >= target_top10):
            print("Total Time: " + str(tot_time) + "s")
            return

        if not wrapper and torch.cuda.is_available():
            del inputs
            del labels
            del outputs
            del predicted
            del top10
            torch.cuda.empty_cache()

    del model

    if verbose:
        print("Total Time: " + str(tot_time) + "s")
